mod opts;

pub mod replacer;
pub mod utils;

use ilhook::x64::Registers;
use replacer::{AbstractReplacer, Replacer};
use std::ffi::{c_char, CStr, CString};
use std::sync::OnceLock;

type CurlEasySetStr = fn(handle: usize, tag: u64, value: *const c_char);

#[derive(Debug)]
pub struct CurlConfig {
    pub handle_rcx_relative_offset: u64,
    pub url_handle_relative_offset: usize,
    pub http_headers_handle_relative_offset: Option<u64>,
    pub curl_easy_setopt: usize,
    pub curl_easy_perform: usize,
}

#[derive(Debug)]
pub struct ProxyConfig {
    pub proxy_url: &'static str,
    pub skip_url_replace: bool,
}

#[derive(Debug)]
pub struct CurlUserData {
    pub set_opt_ptr: CurlEasySetStr,
    pub curl_config: &'static CurlConfig,
    pub replacer: Vec<AbstractReplacer>,
    pub proxy: Option<ProxyConfig>,
}

pub fn hook_curl(
    interceptor: &mut interceptor_rs::Interceptor,
    base_addr: usize,
    config: &'static CurlConfig,
    replacer: Vec<AbstractReplacer>,
    proxy_config: Option<ProxyConfig>,
    lock: &OnceLock<CurlUserData>,
) {
    let set_opt_ptr = base_addr + config.curl_easy_setopt;
    let set_opt_transmute = unsafe { std::mem::transmute::<usize, CurlEasySetStr>(set_opt_ptr) };
    lock.set(CurlUserData {
        set_opt_ptr: set_opt_transmute,
        curl_config: config,
        replacer,
        proxy: proxy_config,
    }).unwrap();
    interceptor
        .attach(
            base_addr + config.curl_easy_perform,
            on_curl_easy_perform,
            Some(lock as *const _ as usize),
        )
        .unwrap();
}

unsafe extern "win64" fn on_curl_easy_perform(reg: *mut Registers, user_data: usize) {
    let data = unsafe {
        (*(user_data as *mut OnceLock<CurlUserData>)).get_mut().unwrap()
    };

    let curl_handle = if data.curl_config.handle_rcx_relative_offset == 0 {
        // If relative offset is 0, it means that rcx already contains the handle.
        unsafe { (*reg).rcx as usize }
    } else {
        // Otherwise it means it comes from another struct, so we need to dereference it.
        unsafe { *(((*reg).rcx + data.curl_config.handle_rcx_relative_offset) as *const usize) }
    };
    let url_ptr =
        unsafe { *((curl_handle + data.curl_config.url_handle_relative_offset) as *const usize) };
    let url = unsafe { CStr::from_ptr(url_ptr as *const i8) }
        .to_str()
        .unwrap();
    println!("[curl_easy_perform] Original URL: {url}");

    if let Some(proxy) = &data.proxy {
        println!(
            "[curl_easy_perform] Setting proxy for url: {}, skip url replace: {}",
            proxy.proxy_url, proxy.skip_url_replace
        );
        let url = CString::new(proxy.proxy_url).unwrap();
        (data.set_opt_ptr)(curl_handle, opts::CURL_OPT_PROXY, url.as_ptr());
        if proxy.skip_url_replace {
            return;
        }
    }

    for replacer in &mut data.replacer {
        if let Ok(result) = replacer.replace(url) {
            // URL Replacement
            println!("[curl_easy_perform] Replacement URL: {result}");
            let url = CString::new(result.as_str()).unwrap();
            (data.set_opt_ptr)(curl_handle, opts::CURL_OPT_URL, url.as_ptr());

            // if let Some(offset) = data.curl_config.http_headers_handle_relative_offset {
            //     let headers_ptr = unsafe { *(((*reg).rcx + offset) as *const usize) };
            //     println!("[curl_easy_perform] Headers URL: {headers_ptr}");
            //     println!("[curl_easy_perform] Headers URL: {}", (*reg).rcx + offset);
            //     // let headers = unsafe { std::slice::from_raw_parts(*(headers_ptr as *const usize) as *const u8, 1000) };
            //     // println!("Headers: {headers:02X?}");
            // }
            return;
        };
    }
    println!("No valid url match found");
}
